/*  bulverded.h - bulverde(PXA27x) display driver
BulVerDD
 */

#ifndef CHARDEV_H
#define CHARDEV_H

#include <linux/ioctl.h>

#define BVDD_VRAM_OFFSET_PXAFB (0)
#define BVDD_VRAM_OFFSET_SRAM (480 * 640 * 2)
#define BVDD_VRAM_OFFSET_USER (BVDD_VRAM_OFFSET_SRAM + (256 * 1024))

#define BVDD_FORMAT_RGB2 1	/* 2-bits/pixel [4 entry palette] */
#define BVDD_FORMAT_RGB4 2	/* 4-bits/pixel [16 entry palette] */
#define BVDD_FORMAT_RGB8 3	/* 8-bits/pixel [256 entry palette] */
#define BVDD_FORMAT_RGB16 4	/* 16-bits/pixel */
#define BVDD_FORMAT_RGB18 5	/* 18-bits/pixel, unpacked */
#define BVDD_FORMAT_RGB18P 6	/* 18-bits/pixel, packed */
#define BVDD_FORMAT_RGB19 7	/* 19-bits/pixel, unpacked */
#define BVDD_FORMAT_RGB19P 8	/* 19-bits/pixel, packed */
#define BVDD_FORMAT_RGB24 9	/* 24-bits/pixel */
#define BVDD_FORMAT_RGB25 10	/* 25-bits/pixel */
#define BVDD_FORMAT_YUV444P 17	/* YCbCr4:4:4 Packed */
#define BVDD_FORMAT_YUV444 18	/* YCbCr4:4:4 Planar */
#define BVDD_FORMAT_YUV422 19	/* YCbCr4:2:2 Planar */
#define BVDD_FORMAT_YUV420 20	/* YCbCr4:2:0 Planar */


typedef struct {
    unsigned char format;
    unsigned short width;
    unsigned short height;
    unsigned short x;
    unsigned short y;
    unsigned long offset;	/* aligned on a 16-byte boundary */
    unsigned long u_offset;	/* aligned on a 16-byte boundary */
    unsigned long v_offset;	/* aligned on a 16-byte boundary */
} bvdd_frame_t;

/* The major device number. We can't rely on dynamic 
 * registration any more, because ioctls need to know 
 * it. */
#define BVDD_MAJOR_NUM 240

/* Set the message of the device driver */
#define BVDD_IOCTL_SETUP_VRAM _IOR(BVDD_MAJOR_NUM, 0, uint32_t)
#define BVDD_IOCTL_SETUP_BASE_FRAME _IOR(BVDD_MAJOR_NUM, 1, bvdd_frame_t *)
#define BVDD_IOCTL_SETUP_OVERLAY1_FRAME _IOR(BVDD_MAJOR_NUM, 2, bvdd_frame_t *)
#define BVDD_IOCTL_SETUP_OVERLAY2_FRAME _IOR(BVDD_MAJOR_NUM, 3, bvdd_frame_t *)
#define BVDD_IOCTL_BRANCH_OVERLAY1_FRAME _IOR(BVDD_MAJOR_NUM, 5, bvdd_frame_t *)
#define BVDD_IOCTL_BRANCH_OVERLAY2_FRAME _IOR(BVDD_MAJOR_NUM, 6, bvdd_frame_t *)
#define BVDD_IOCTL_SET_IWMMXT_STATE _IOR(BVDD_MAJOR_NUM, 7, int)
#define BVDD_IOCTL_GET_CCCR _IOR(BVDD_MAJOR_NUM, 11, int *)
#define BVDD_IOCTL_SET_CCCR _IOR(BVDD_MAJOR_NUM, 12, int)
#define BVDD_IOCTL_GET_VCORE _IOR(BVDD_MAJOR_NUM, 13, int *)
#define BVDD_IOCTL_SET_VCORE _IOR(BVDD_MAJOR_NUM, 14, int)

#define BVDD_IOCTL_CHANGE_QVGA _IO(BVDD_MAJOR_NUM, 15)
#define BVDD_IOCTL_CHANGE_VGA  _IO(BVDD_MAJOR_NUM, 16)

#define BVDD_IOCTL_DEBUG _IO(BVDD_MAJOR_NUM, 10)

/* The name of the device file */
#define BVDD_DEVICE_FILE_NAME "bvdd"


#endif
